@@ -26,42 +26,49 @@
 #include "dbus.h"
 #include "util.h"
 #include "imstatus.h"
+#include "vgl-bookmark-mgr.h"
+#include "vgl-bookmark-window.h"
 
 #ifdef HAVE_TRAY_ICON
-#include "vagalume-tray-icon.h"
+#include "vgl-tray-icon.h"
 #endif
 
-static lastfm_session *session = NULL;
-static lastfm_pls *playlist = NULL;
-static rsp_session *rsp_sess = NULL;
-static lastfm_mainwin *mainwin = NULL;
-static lastfm_usercfg *usercfg = NULL;
+#ifdef MAEMO
+#include <libosso.h>
+#endif
+
+static LastfmSession *session = NULL;
+static LastfmPls *playlist = NULL;
+static RspSession *rsp_sess = NULL;
+static VglMainWindow *mainwin = NULL;
+static VglUserCfg *usercfg = NULL;
 static GList *friends = NULL;
 static GList *usertags = NULL;
-static lastfm_track *nowplaying = NULL;
+static LastfmTrack *nowplaying = NULL;
 static time_t nowplaying_since = 0;
-static rsp_rating nowplaying_rating = RSP_RATING_NONE;
+static RspRating nowplaying_rating = RSP_RATING_NONE;
 static gboolean showing_cover = FALSE;
 static gboolean stopping_after_track = FALSE;
+static gboolean shutting_down = FALSE;
 
 #ifdef HAVE_TRAY_ICON
-static VagalumeTrayIcon *tray_icon = NULL;
+static VglTrayIcon *tray_icon = NULL;
 #endif
 
 typedef struct {
-        lastfm_track *track;
-        rsp_rating rating;
+        LastfmTrack *track;
+        RspRating rating;
         time_t start;
 } rsp_data;
 
 typedef struct {
-        lastfm_track *track;
+        LastfmTrack *track;
         char *taglist;                /* comma-separated list of tags */
         request_type type;
 } tag_data;
 
 typedef struct {
-        lastfm_track *track;
+        LastfmTrack *track;
         char *rcpt;                  /* Recipient of the recommendation */
         char *text;                  /* text of the recommendation */
         request_type type;
@@ -90,8 +97,8 @@ typedef struct {
 void
 controller_show_error(const char *text)
 {
-        g_return_if_fail(mainwin != NULL);
-        ui_error_dialog(mainwin_get_window(mainwin, TRUE), text);
+        g_return_if_fail(VGL_IS_MAIN_WINDOW(mainwin));
+        ui_error_dialog(vgl_main_window_get_window(mainwin, TRUE), text);
 }
 
 /**
@@ -102,8 +109,8 @@ controller_show_error(const char *text)
 void
 controller_show_warning(const char *text)
 {
-        g_return_if_fail(mainwin != NULL);
-        ui_warning_dialog(mainwin_get_window(mainwin, TRUE), text);
+        g_return_if_fail(VGL_IS_MAIN_WINDOW(mainwin));
+        ui_warning_dialog(vgl_main_window_get_window(mainwin, TRUE), text);
 }
 
 /**
@@ -114,8 +121,8 @@ controller_show_warning(const char *text)
 void
 controller_show_info(const char *text)
 {
-        g_return_if_fail(mainwin != NULL);
-        ui_info_dialog(mainwin_get_window(mainwin, TRUE), text);
+        g_return_if_fail(VGL_IS_MAIN_WINDOW(mainwin));
+        ui_info_dialog(vgl_main_window_get_window(mainwin, TRUE), text);
 }
 
 /**
@@ -127,9 +134,9 @@ void
 controller_show_banner(const char *text)
 {
 #ifdef MAEMO
-        ui_info_banner(mainwin_get_window(mainwin, FALSE), text);
+        ui_info_banner(vgl_main_window_get_window(mainwin, FALSE), text);
 #else
-        mainwin_show_status_msg(mainwin, text);
+        vgl_main_window_show_status_msg(mainwin, text);
 #endif
 }
 
@@ -142,12 +149,13 @@ controller_show_banner(const char *text)
 gboolean
 controller_confirm_dialog(const char *text, gboolean show_always)
 {
-        g_return_val_if_fail(mainwin != NULL, FALSE);
+        g_return_val_if_fail(VGL_IS_MAIN_WINDOW(mainwin), FALSE);
 
         if (!show_always && usercfg->disable_confirm_dialogs)
                 return TRUE;
 
-        return ui_confirm_dialog(mainwin_get_window(mainwin, FALSE), text);
+        return ui_confirm_dialog(vgl_main_window_get_window(mainwin, FALSE),
+                                 text);
 }
 
 /**
@@ -158,8 +166,8 @@ controller_confirm_dialog(const char *text, gboolean show_always)
 void
 controller_show_mainwin(gboolean show)
 {
-        g_return_if_fail(mainwin != NULL);
-        mainwin_show_window(mainwin, show);
+        g_return_if_fail(VGL_IS_MAIN_WINDOW(mainwin));
+        vgl_main_window_show(mainwin, show);
 }
 
 /**
@@ -170,8 +178,8 @@ controller_show_mainwin(gboolean show)
 void
 controller_toggle_mainwin_visibility (void)
 {
-        g_return_if_fail(mainwin != NULL);
-        mainwin_toggle_visibility (mainwin);
+        g_return_if_fail(VGL_IS_MAIN_WINDOW(mainwin));
+        vgl_main_window_toggle_visibility(mainwin);
 }
 
 /**
@@ -179,19 +187,19 @@ controller_toggle_mainwin_visibility (void)
  * updates the UI (progressbars, etc). to reflect that. To be called
  * using g_timeout_add()
  *
- * @param data A pointer to the lastfm_track being played
+ * @param data A pointer to the LastfmTrack being played
  * @return TRUE if the track hasn't fininished yet, FALSE otherwise
  */
 static gboolean
 controller_show_progress(gpointer data)
 {
-        lastfm_track *tr = (lastfm_track *) data;
-        g_return_val_if_fail(mainwin != NULL && tr != NULL, FALSE);
+        LastfmTrack *tr = (LastfmTrack *) data;
+        g_return_val_if_fail(VGL_IS_MAIN_WINDOW(mainwin) && tr, FALSE);
         if (nowplaying != NULL && tr->id == nowplaying->id) {
                 guint played = lastfm_audio_get_running_time();
                 if (played != -1) {
                         guint length = nowplaying->duration/1000;
-                        mainwin_show_progress(mainwin, length, played);
+                        vgl_main_window_show_progress(mainwin,length,played);
                 }
                 return TRUE;
         } else {
@@ -204,7 +212,7 @@ controller_show_progress(gpointer data)
  * Gets the track that is currently being played
  * @return The track
  */
-lastfm_track *
+LastfmTrack *
 controller_get_current_track(void)
 {
   return nowplaying;
@@ -251,7 +259,7 @@ set_user_tag_list(const char *user, GList *list)
  * @param sess The new session (can be NULL)
  */
 static void
-set_rsp_session(const char *user, rsp_session *sess)
+set_rsp_session(const char *user, RspSession *sess)
 {
         g_return_if_fail(user != NULL && usercfg != NULL);
         if (!strcmp(user, usercfg->username)) {
@@ -260,6 +268,34 @@ set_rsp_session(const char *user, rsp_session *sess)
         }
 }
 
+/**
+ * Renews the RSP session. Must be called from a thread.
+ */
+static void
+renew_rsp_session (void)
+{
+        char *user, *pass;
+        RspSession *session;
+
+        g_return_if_fail (usercfg != NULL);
+
+        gdk_threads_enter ();
+        user = g_strdup (usercfg->username);
+        pass = g_strdup (usercfg->password);
+        gdk_threads_leave ();
+
+        g_debug ("Renewing RSP session ...");
+        session = rsp_session_new (user, pass, NULL);
+        if (session != NULL) {
+                gdk_threads_enter ();
+                set_rsp_session (user, session);
+                gdk_threads_leave ();
+        }
+
+        g_free (user);
+        g_free (pass);
+}
+
 /**
  * Scrobble a track using the Audioscrobbler Realtime Submission
  * Protocol. This can take some seconds, so it must be called using
@@ -273,7 +309,7 @@ static gpointer
 scrobble_track_thread(gpointer data)
 {
         rsp_data *d = (rsp_data *) data;
-        rsp_session *s = NULL;
+        RspSession *s = NULL;
         g_return_val_if_fail(d != NULL && d->track != NULL && d->start > 0,
                              NULL);
         char *user = NULL, *pass = NULL;
@@ -285,8 +321,17 @@ scrobble_track_thread(gpointer data)
         }
         gdk_threads_leave();
         if (s != NULL) {
-                rsp_scrobble(s, d->track, d->start, d->rating);
-                rsp_session_destroy(s);
+                RspResponse ret;
+                ret = rsp_scrobble (s, d->track, d->start, d->rating);
+                if (ret == RSP_RESPONSE_BADSESSION) {
+                        rsp_session_destroy (s);
+                        renew_rsp_session ();
+                        gdk_threads_enter ();
+                        s = rsp_session_copy (rsp_sess);
+                        gdk_threads_leave ();
+                        rsp_scrobble (s, d->track, d->start, d->rating);
+                }
+                rsp_session_destroy (s);
         }
         /* This love_ban_track() call won't be needed anymore with
          * Lastfm's new protocol v1.2 */
@@ -345,7 +390,7 @@ set_nowplaying_thread(gpointer data)
 {
         rsp_data *d = (rsp_data *) data;
         g_return_val_if_fail(d != NULL && d->track != NULL, FALSE);
-        rsp_session *s = NULL;
+        RspSession *s = NULL;
         gboolean set_np = FALSE;
         g_usleep(10 * G_USEC_PER_SEC);
         gdk_threads_enter();
@@ -355,8 +400,16 @@ set_nowplaying_thread(gpointer data)
         }
         gdk_threads_leave();
         if (set_np) {
-                rsp_set_nowplaying(s, d->track);
-                rsp_session_destroy(s);
+                RspResponse ret = rsp_set_nowplaying (s, d->track);
+                if (ret == RSP_RESPONSE_BADSESSION) {
+                        rsp_session_destroy (s);
+                        renew_rsp_session ();
+                        gdk_threads_enter ();
+                        s = rsp_session_copy (rsp_sess);
+                        gdk_threads_leave ();
+                        rsp_set_nowplaying (s, d->track);
+                }
+                rsp_session_destroy (s);
         }
         lastfm_track_unref(d->track);
         g_slice_free(rsp_data, d);
@@ -373,7 +426,7 @@ set_nowplaying_thread(gpointer data)
  * @param track The track to be set as Now Playing
  */
 static void
-controller_set_nowplaying(lastfm_track *track)
+controller_set_nowplaying(LastfmTrack *track)
 {
         g_return_if_fail(usercfg != NULL);
         if (nowplaying != NULL) {
@@ -405,7 +458,7 @@ get_user_extradata(void)
         gboolean rsp_ok = FALSE;
         gboolean friends_ok = FALSE;
         gboolean usertags_ok = FALSE;
-        rsp_session *sess = NULL;
+        RspSession *sess = NULL;
         GList *friends = NULL;
         GList *usertags = NULL;
         gdk_threads_enter();
@@ -483,7 +536,7 @@ apply_usercfg(void)
         }
 #ifdef HAVE_TRAY_ICON
         if (tray_icon != NULL) {
-                vagalume_tray_icon_show_notifications (
+                vgl_tray_icon_show_notifications (
                         tray_icon, usercfg->show_notifications);
         }
 #endif
@@ -497,7 +550,7 @@ apply_usercfg(void)
 void
 controller_open_usercfg(void)
 {
-        g_return_if_fail(mainwin != NULL);
+        g_return_if_fail(VGL_IS_MAIN_WINDOW(mainwin));
         gboolean userchanged = FALSE;
         gboolean pwchanged = FALSE;
         gboolean changed;
@@ -506,11 +559,11 @@ controller_open_usercfg(void)
         char *oldpw = usercfg != NULL ? g_strdup(usercfg->password) :
                                         g_strdup("");
 
-        changed = ui_usercfg_window(mainwin_get_window(mainwin, FALSE),
-                                    &usercfg);
+        changed = ui_usercfg_window(
+                vgl_main_window_get_window(mainwin, FALSE), &usercfg);
 
         if (changed && usercfg != NULL) {
-                write_usercfg(usercfg);
+                vgl_user_cfg_write(usercfg);
                 userchanged = strcmp(olduser, usercfg->username);
                 pwchanged = strcmp(oldpw, usercfg->password);
                 apply_usercfg();
@@ -538,7 +591,7 @@ controller_open_usercfg(void)
 static gboolean
 check_usercfg(gboolean ask)
 {
-        if (usercfg == NULL) usercfg = read_usercfg();
+        if (usercfg == NULL) usercfg = vgl_user_cfg_read();
         if (usercfg == NULL && ask) controller_open_usercfg();
         if (usercfg != NULL) apply_usercfg();
         return (usercfg != NULL);
@@ -562,8 +615,8 @@ check_session_thread(gpointer userdata)
         g_return_val_if_fail(userdata != NULL, NULL);
         check_session_thread_data *data;
         gboolean connected = FALSE;
-        lastfm_err err = LASTFM_ERR_NONE;
-        lastfm_session *s;
+        LastfmErr err = LASTFM_ERR_NONE;
+        LastfmSession *s;
         data = (check_session_thread_data *) userdata;
         s = lastfm_session_new(data->user, data->pass, &err);
         if (s == NULL || s->id == NULL) {
@@ -579,7 +632,8 @@ check_session_thread(gpointer userdata)
         } else {
                 gdk_threads_enter();
                 session = s;
-                mainwin_set_ui_state(mainwin, LASTFM_UI_STATE_STOPPED, NULL);
+                vgl_main_window_set_state(mainwin,
+                                          VGL_MAIN_WINDOW_STATE_STOPPED,NULL);
                 gdk_threads_leave();
                 connected = TRUE;
         }
@@ -640,8 +694,8 @@ check_session(check_session_cb success_cb, check_session_cb failure_cb,
                         data->failure_cb = failure_cb;
                         data->cbdata = cbdata;
                         connection_go_online(check_session_conn_cb, data);
-                        mainwin_set_ui_state(mainwin,
-                                             LASTFM_UI_STATE_CONNECTING,
+                        vgl_main_window_set_state(mainwin,
+                                             VGL_MAIN_WINDOW_STATE_CONNECTING,
                                              NULL);
                 } else {
                         controller_disconnect();
@@ -657,19 +711,19 @@ check_session(check_session_cb success_cb, check_session_cb failure_cb,
 /**
  * Set the album cover image. This must be done in a thread to avoid
  * freezing the UI.
- * @param data A pointer to the lastfm_track
+ * @param data A pointer to the LastfmTrack
  * @return NULL (not used)
  */
 static gpointer
 set_album_cover_thread(gpointer data)
 {
-        lastfm_track *t = (lastfm_track *) data;
+        LastfmTrack *t = (LastfmTrack *) data;
         g_return_val_if_fail(t != NULL && t->image_url != NULL, NULL);
         lastfm_get_track_cover_image(t);
         if (t->image_data == NULL) g_warning("Error getting cover image");
         gdk_threads_enter();
         if (mainwin && nowplaying == t) {
-                mainwin_set_album_cover(mainwin, t->image_data,
+                vgl_main_window_set_album_cover(mainwin, t->image_data,
                                         t->image_data_size);
         }
         gdk_threads_leave();
@@ -683,16 +737,16 @@ set_album_cover_thread(gpointer data)
  * otherwise. This can take a bit, so it is done in a separate thread
  * to avoid freezing the UI.
  *
- * @param data A copy of the lastfm_session used to get the new
+ * @param data A copy of the LastfmSession used to get the new
  *             playlist.
  * @return NULL (not used)
  */
 static gpointer
 start_playing_get_pls_thread(gpointer data)
 {
-        lastfm_session *s = (lastfm_session *) data;
+        LastfmSession *s = (LastfmSession *) data;
         g_return_val_if_fail(s != NULL && usercfg != NULL, NULL);
-        lastfm_pls *pls = lastfm_request_playlist(s, usercfg->discovery_mode);
+        LastfmPls *pls = lastfm_request_playlist(s, usercfg->discovery_mode);
         gdk_threads_enter();
         if (pls == NULL) {
                 controller_stop_playing();
@@ -717,14 +771,15 @@ start_playing_get_pls_thread(gpointer data)
 void
 controller_show_cover(void)
 {
-        g_return_if_fail(mainwin != NULL);
-        if (showing_cover || nowplaying == NULL || mainwin->is_hidden) return;
+        g_return_if_fail(VGL_IS_MAIN_WINDOW(mainwin));
+        if (showing_cover || nowplaying == NULL ||
+            vgl_main_window_is_hidden(mainwin)) return;
         showing_cover = TRUE;
         if (nowplaying->image_url != NULL) {
                 g_thread_create(set_album_cover_thread,
                                 lastfm_track_ref(nowplaying), FALSE, NULL);
         } else {
-                mainwin_set_album_cover(mainwin, NULL, 0);
+                vgl_main_window_set_album_cover(mainwin, NULL, 0);
         }
 }
 
@@ -735,10 +790,11 @@ controller_show_cover(void)
 static void
 controller_audio_started_cb(void)
 {
-        g_return_if_fail(mainwin != NULL && nowplaying != NULL);
-        lastfm_track *track;
+        g_return_if_fail(VGL_IS_MAIN_WINDOW(mainwin) && nowplaying);
+        LastfmTrack *track;
         nowplaying_since = time(NULL);
-        mainwin_set_ui_state(mainwin, LASTFM_UI_STATE_PLAYING, nowplaying);
+        vgl_main_window_set_state(mainwin, VGL_MAIN_WINDOW_STATE_PLAYING,
+                                  nowplaying);
         track = lastfm_track_ref(nowplaying);
         controller_show_progress(track);
         im_set_status(usercfg, track);
@@ -758,11 +814,12 @@ controller_audio_started_cb(void)
 static void
 controller_start_playing_cb(gpointer userdata)
 {
-        lastfm_track *track = NULL;
+        LastfmTrack *track = NULL;
         g_return_if_fail(mainwin && playlist && nowplaying == NULL);
-        mainwin_set_ui_state(mainwin, LASTFM_UI_STATE_CONNECTING, NULL);
+        vgl_main_window_set_state(mainwin, VGL_MAIN_WINDOW_STATE_CONNECTING,
+                                  NULL);
         if (lastfm_pls_size(playlist) == 0) {
-                lastfm_session *s = lastfm_session_copy(session);
+                LastfmSession *s = lastfm_session_copy(session);
                 g_thread_create(start_playing_get_pls_thread,s,FALSE,NULL);
                 return;
         }
@@ -774,7 +831,7 @@ controller_start_playing_cb(gpointer userdata)
 
 #ifdef HAVE_TRAY_ICON
         if (tray_icon) {
-                vagalume_tray_icon_notify_playback (tray_icon, track);
+                vgl_tray_icon_notify_playback (tray_icon, track);
         }
 #endif
 
@@ -809,12 +866,13 @@ controller_start_playing(void)
 
 /**
  * Stop the track being played (if any) and scrobbles it. To be called
- * only from controller_stop_playing() and controller_skip_track()
+ * only from controller_stop_playing(), controller_skip_track() and
+ * controller_play_radio_by_url()
  */
 static void
 finish_playing_track(void)
 {
-        g_return_if_fail(mainwin != NULL);
+        g_return_if_fail(VGL_IS_MAIN_WINDOW(mainwin));
         if (nowplaying != NULL) {
                 controller_scrobble_track();
                 controller_set_nowplaying(NULL);
@@ -828,11 +886,16 @@ finish_playing_track(void)
 void
 controller_stop_playing(void)
 {
-        g_return_if_fail(mainwin != NULL);
-        lastfm_ui_state new_state = session != NULL ?
-                LASTFM_UI_STATE_STOPPED :
-                LASTFM_UI_STATE_DISCONNECTED;
-        mainwin_set_ui_state(mainwin, new_state, NULL);
+        g_return_if_fail(VGL_IS_MAIN_WINDOW(mainwin));
+        VglMainWindowState new_state = session != NULL ?
+                VGL_MAIN_WINDOW_STATE_STOPPED :
+                VGL_MAIN_WINDOW_STATE_DISCONNECTED;
+
+        /* Updating the window title just before destroying the window
+         * causes a crash, at least in the Moblin platform */
+        if (!shutting_down)
+                vgl_main_window_set_state (mainwin, new_state, NULL);
+
         finish_playing_track();
         stopping_after_track = FALSE;
         im_clear_status();
@@ -842,7 +905,7 @@ controller_stop_playing(void)
 
 #ifdef HAVE_TRAY_ICON
         if (tray_icon) {
-                vagalume_tray_icon_notify_playback (tray_icon, NULL);
+                vgl_tray_icon_notify_playback (tray_icon, NULL);
         }
 #endif
 }
@@ -853,7 +916,7 @@ controller_stop_playing(void)
 void
 controller_skip_track(void)
 {
-        g_return_if_fail(mainwin != NULL);
+        g_return_if_fail(VGL_IS_MAIN_WINDOW(mainwin));
         finish_playing_track();
         controller_start_playing();
 }
@@ -883,7 +946,7 @@ void
 controller_download_track(void)
 {
         g_return_if_fail(nowplaying && nowplaying->free_track_url && usercfg);
-        lastfm_track *t = lastfm_track_ref(nowplaying);
+        LastfmTrack *t = lastfm_track_ref(nowplaying);
         if (controller_confirm_dialog(_("Download this track?"), FALSE)) {
                 char *filename, *dstpath;
                 gboolean download = TRUE;
@@ -907,6 +970,52 @@ controller_download_track(void)
         }
 }
 
+/**
+ * Opens the bookmarks window
+ */
+void
+controller_manage_bookmarks(void)
+{
+        g_return_if_fail(VGL_IS_MAIN_WINDOW(mainwin));
+        vgl_bookmark_window_show(vgl_main_window_get_window(mainwin, FALSE));
+}
+
+/*
+ * Bookmark the current track
+ */
+void controller_add_bookmark(request_type type)
+{
+        g_return_if_fail(nowplaying != NULL && VGL_IS_MAIN_WINDOW(mainwin));
+        char *name, *url;
+        const char *banner;
+        VglBookmarkMgr *mgr = vgl_bookmark_mgr_get_instance();
+        if (type == REQUEST_ARTIST) {
+                g_return_if_fail(nowplaying->artistid != 0);
+                name = g_strdup(nowplaying->artist);
+                url = g_strdup_printf("lastfm://play/artists/%u",
+                                      nowplaying->artistid);
+                banner = _("Artist added to bookmarks");
+        } else if (type == REQUEST_TRACK) {
+                g_return_if_fail(nowplaying->id != 0);
+                name = g_strdup_printf("%s - '%s'", nowplaying->artist,
+                                       nowplaying->title);
+                url = g_strdup_printf("lastfm://play/tracks/%u",
+                                      nowplaying->id);
+                banner = _("Track added to bookmarks");
+        } else {
+                g_critical("Bookmark request not supported");
+                return;
+        }
+        if (ui_edit_bookmark_dialog(vgl_main_window_get_window(mainwin, FALSE),
+                                    &name, &url, TRUE)) {
+                vgl_bookmark_mgr_add_bookmark(mgr, name, url);
+                vgl_bookmark_mgr_save_to_disk (mgr, TRUE);
+                controller_show_banner(banner);
+        }
+        g_free(name);
+        g_free(url);
+}
+
 /**
  * Tells Vagalume to stop (or not) after the current track
  * @param stop Whether to stop or not
@@ -925,11 +1034,11 @@ controller_set_stop_after(gboolean stop)
 void
 controller_love_track(gboolean interactive)
 {
-        g_return_if_fail(nowplaying != NULL && mainwin != NULL);
+        g_return_if_fail(nowplaying != NULL && VGL_IS_MAIN_WINDOW(mainwin));
         if (!interactive ||
             controller_confirm_dialog(_("Really mark track as loved?"),FALSE)) {
                 nowplaying_rating = RSP_RATING_LOVE;
-                mainwin_set_track_as_loved(mainwin);
+                vgl_main_window_set_track_as_loved(mainwin);
                 if (interactive) {
                         controller_show_banner(_("Marking track as loved"));
                 }
@@ -1013,16 +1122,19 @@ controller_tag_track()
         /* Keep this static to remember the previous value */
         static request_type type = REQUEST_ARTIST;
         char *tags = NULL;
-        lastfm_track *track = lastfm_track_ref(nowplaying);
+        LastfmTrack *track = lastfm_track_ref(nowplaying);
         gboolean accept;
         if (track->album[0] == '\0' && type == REQUEST_ALBUM) {
                 type = REQUEST_ARTIST;
         }
-        accept = tagwin_run(mainwin_get_window(mainwin, FALSE),
+        accept = tagwin_run(vgl_main_window_get_window(mainwin, FALSE),
                             usercfg->username, &tags,
                             usertags, track, &type);
-        if (accept && tags != NULL && tags[0] != '\0') {
+        if (accept) {
                 tag_data *d = g_slice_new0(tag_data);
+                if (tags == NULL) {
+                        tags = g_strdup("");
+                }
                 d->track = track;
                 d->taglist = tags;
                 d->type = type;
@@ -1088,13 +1200,13 @@ controller_recomm_track(void)
         char *body = NULL;
         /* Keep this static to remember the previous value */
         static request_type type = REQUEST_TRACK;
-        lastfm_track *track = lastfm_track_ref(nowplaying);
+        LastfmTrack *track = lastfm_track_ref(nowplaying);
         gboolean accept;
         if (track->album[0] == '\0' && type == REQUEST_ALBUM) {
                 type = REQUEST_ARTIST;
         }
 
-        accept = recommwin_run(mainwin_get_window(mainwin, FALSE),
+        accept = recommwin_run(vgl_main_window_get_window(mainwin, FALSE),
                                &rcpt, &body, friends, track, &type);
         if (accept && rcpt && body && rcpt[0] && body[0]) {
                 g_strstrip(rcpt);
@@ -1119,14 +1231,14 @@ controller_recomm_track(void)
  * Add a track to the user's playlist. This can take some seconds, so
  * it must be called using g_thread_create() to avoid freezing the UI.
  *
- * @param data Pointer to the lastfm_track to add. This data must be
+ * @param data Pointer to the LastfmTrack to add. This data must be
  *             freed here
  * @return NULL (this value is not used)
  */
 gpointer
 add_to_playlist_thread(gpointer data)
 {
-        lastfm_track *t = (lastfm_track *) data;
+        LastfmTrack *t = (LastfmTrack *) data;
         g_return_val_if_fail(t != NULL, NULL);
         gboolean retval = FALSE;
         char *user = NULL, *pass = NULL;
@@ -1146,7 +1258,8 @@ add_to_playlist_thread(gpointer data)
                 controller_show_banner(retval ?
                                        _("Track added to playlist") :
                                        _("Error adding track to playlist"));
-                mainwin_set_track_as_added_to_playlist(mainwin, retval);
+                vgl_main_window_set_track_as_added_to_playlist(mainwin,
+                                                               retval);
         }
         gdk_threads_leave();
         lastfm_track_unref(t);
@@ -1163,8 +1276,8 @@ controller_add_to_playlist(void)
         g_return_if_fail(usercfg != NULL && nowplaying != NULL);
         if (controller_confirm_dialog(
                     _("Really add this track to the playlist?"), FALSE)) {
-                lastfm_track *track = lastfm_track_ref(nowplaying);
-                mainwin_set_track_as_added_to_playlist(mainwin, TRUE);
+                LastfmTrack *track = lastfm_track_ref(nowplaying);
+                vgl_main_window_set_track_as_added_to_playlist(mainwin, TRUE);
                 g_thread_create(add_to_playlist_thread,track,FALSE,NULL);
         }
 }
@@ -1173,18 +1286,31 @@ controller_add_to_playlist(void)
  * Start playing a radio by its URL, stopping the current track if
  * necessary
  *
- * This is the success callback of controller_play_radio_by_url()
+ * This is a thread created from controller_play_radio_by_url_cb()
  *
- * @param url The URL of the radio to be played (freed by this function)
+ * @param data The URL of the radio to be played (freed by this function)
+ * @return NULL (not used)
  */
-static void
-controller_play_radio_by_url_cb(char *url)
+static gpointer
+controller_play_radio_by_url_thread(gpointer data)
 {
+        char *url = (char *) data;
+        LastfmSession *sess;
+        gdk_threads_enter();
+        if (!VGL_IS_MAIN_WINDOW(mainwin) || session == NULL) {
+                g_critical("Main window destroyed or session not found");
+                gdk_threads_leave();
+                return NULL;
+        }
+        sess = lastfm_session_copy(session);
         if (url == NULL) {
                 g_critical("Attempted to play a NULL radio URL");
                 controller_stop_playing();
         } else if (lastfm_radio_url_is_custom(url)) {
-                lastfm_pls *pls = lastfm_request_custom_playlist(session, url);
+                LastfmPls *pls;
+                gdk_threads_leave();
+                pls = lastfm_request_custom_playlist(sess, url);
+                gdk_threads_enter();
                 if (pls != NULL) {
                         lastfm_pls_destroy(playlist);
                         playlist = pls;
@@ -1193,17 +1319,42 @@ controller_play_radio_by_url_cb(char *url)
                         controller_stop_playing();
                         controller_show_info(_("Invalid radio URL"));
                 }
-        } else if (lastfm_set_radio(session, url)) {
-                lastfm_pls_clear(playlist);
-                controller_skip_track();
         } else {
-                controller_stop_playing();
-                controller_show_info(_("Invalid radio URL. Either\n"
-                                       "this radio doesn't exist\n"
-                                       "or it is only available\n"
-                                       "for Last.fm subscribers"));
+                gboolean radio_set;
+                gdk_threads_leave();
+                radio_set = lastfm_set_radio(sess, url);
+                gdk_threads_enter();
+                if (radio_set) {
+                        lastfm_pls_clear(playlist);
+                        controller_skip_track();
+                } else {
+                        controller_stop_playing();
+                        controller_show_info(_("Invalid radio URL. Either\n"
+                                               "this radio doesn't exist\n"
+                                               "or it is only available\n"
+                                               "for Last.fm subscribers"));
+                }
         }
+        gdk_threads_leave();
         g_free(url);
+        lastfm_session_destroy(sess);
+        return NULL;
+}
+
+/**
+ * Start playing a radio by its URL, stopping the current track if
+ * necessary
+ *
+ * This is the success callback of controller_play_radio_by_url().
+ * This just creates a thread, the actual code is in
+ * controller_play_radio_by_url_thread()
+ *
+ * @param url The URL of the radio to be played (freed by this function)
+ */
+static void
+controller_play_radio_by_url_cb(char *url)
+{
+        g_thread_create(controller_play_radio_by_url_thread, url, FALSE, NULL);
 }
 
 /**
@@ -1218,6 +1369,9 @@ void
 controller_play_radio_by_url(const char *url)
 {
         check_session_cb cb;
+        finish_playing_track();
+        vgl_main_window_set_state(mainwin, VGL_MAIN_WINDOW_STATE_CONNECTING,
+                                  NULL);
         cb = (check_session_cb) controller_play_radio_by_url_cb;
         check_session(cb, (check_session_cb) g_free, g_strdup(url));
 }
@@ -1241,11 +1395,10 @@ controller_play_radio_cb(gpointer userdata)
         } else if (type == LASTFM_USERTAG_RADIO) {
                 static char *previous = NULL;
                 char *tag;
-                tag = ui_input_dialog_with_list(mainwin_get_window(mainwin,
-                                                                   TRUE),
-                                                _("Enter tag"),
-                                                _("Enter one of your tags"),
-                                                usertags, previous);
+                tag = ui_input_dialog_with_list(
+                        vgl_main_window_get_window(mainwin, TRUE),
+                        _("Enter tag"), _("Enter one of your tags"),
+                        usertags, previous);
                 if (tag != NULL) {
                         url = lastfm_usertag_radio_url(usercfg->username, tag);
                         /* Store the new value for later use */
@@ -1292,7 +1445,7 @@ controller_play_others_radio_cb(gpointer userdata)
         lastfm_radio type = GPOINTER_TO_INT(userdata);
         static char *previous = NULL;
         char *url = NULL;
-        char *user = ui_input_dialog_with_list(mainwin_get_window(mainwin,
+        char *user = ui_input_dialog_with_list(vgl_main_window_get_window(mainwin,
                                                                   TRUE),
                                                _("Enter user name"),
                                                _("Play this user's radio"),
@@ -1331,12 +1484,12 @@ controller_play_others_radio(lastfm_radio type)
 void
 controller_play_group_radio(void)
 {
-        g_return_if_fail(mainwin != NULL);
+        g_return_if_fail(VGL_IS_MAIN_WINDOW(mainwin));
         static char *previous = NULL;
         char *url = NULL;
-        char *group = ui_input_dialog(mainwin_get_window(mainwin, TRUE),
-                                      _("Enter group"),
-                                      _("Enter group name"), previous);
+        char *group = ui_input_dialog(
+                vgl_main_window_get_window(mainwin, TRUE),
+                _("Enter group"), _("Enter group name"), previous);
         if (group != NULL) {
                 url = lastfm_radio_url(LASTFM_GROUP_RADIO, group);
                 controller_play_radio_by_url(url);
@@ -1353,14 +1506,13 @@ controller_play_group_radio(void)
 void
 controller_play_globaltag_radio(void)
 {
-        g_return_if_fail(mainwin != NULL);
+        g_return_if_fail(VGL_IS_MAIN_WINDOW(mainwin));
         static char *previous = NULL;
         char *url = NULL;
         char *tag;
-        tag = ui_input_dialog_with_list(mainwin_get_window(mainwin, TRUE),
-                                        _("Enter tag"),
-                                        _("Enter a global tag"),
-                                        usertags, previous);
+        tag = ui_input_dialog_with_list(
+                vgl_main_window_get_window(mainwin, TRUE),
+                _("Enter tag"), _("Enter a global tag"), usertags, previous);
         if (tag != NULL) {
                 url = lastfm_radio_url(LASTFM_GLOBALTAG_RADIO, tag);
                 controller_play_radio_by_url(url);
@@ -1379,12 +1531,12 @@ controller_play_globaltag_radio(void)
 void
 controller_play_similarartist_radio(void)
 {
-        g_return_if_fail(mainwin != NULL);
+        g_return_if_fail(VGL_IS_MAIN_WINDOW(mainwin));
         static char *previous = NULL;
         char *url = NULL;
-        char *artist = ui_input_dialog(mainwin_get_window(mainwin, TRUE),
-                                       _("Enter artist"),
-                                       _("Enter an artist's name"), previous);
+        char *artist = ui_input_dialog(
+                vgl_main_window_get_window(mainwin, TRUE),
+                _("Enter artist"), _("Enter an artist's name"), previous);
         if (artist != NULL) {
                 url = lastfm_radio_url(LASTFM_SIMILAR_ARTIST_RADIO, artist);
                 controller_play_radio_by_url(url);
@@ -1401,10 +1553,10 @@ controller_play_similarartist_radio(void)
 void
 controller_play_radio_ask_url(void)
 {
-        g_return_if_fail(mainwin != NULL);
+        g_return_if_fail(VGL_IS_MAIN_WINDOW(mainwin));
         static char *previous = NULL;
         char *url = NULL;
-        url = ui_input_dialog(mainwin_get_window(mainwin, TRUE),
+        url = ui_input_dialog(vgl_main_window_get_window(mainwin, TRUE),
                               _("Enter radio URL"),
                               _("Enter the URL of the Last.fm radio"),
                               previous ? previous : "lastfm://");
@@ -1455,8 +1607,10 @@ controller_set_volume(int vol)
 void
 controller_quit_app(void)
 {
+        g_return_if_fail(VGL_IS_MAIN_WINDOW(mainwin));
+        shutting_down = TRUE;
         controller_stop_playing();
-        mainwin_quit_app();
+        vgl_main_window_destroy(mainwin);
 }
 
 /**
@@ -1469,23 +1623,50 @@ controller_quit_app(void)
  *                  command line
  */
 void
-controller_run_app(lastfm_mainwin *win, const char *radio_url)
+controller_run_app (const char *radio_url)
 {
-        g_return_if_fail(win != NULL && mainwin == NULL);
-        const char *errmsg;
-        mainwin = win;
-        mainwin_show_window(mainwin, TRUE);
-        mainwin_set_ui_state(mainwin, LASTFM_UI_STATE_DISCONNECTED, NULL);
+        g_return_if_fail (mainwin == NULL);
+        const char *errmsg = NULL;
+#ifdef MAEMO
+        osso_context_t *osso_context = NULL;
+#endif
+        DbusInitReturnCode dbuscode = lastfm_dbus_init ();
+        if (dbuscode == DBUS_INIT_ERROR) {
+                errmsg = _("Unable to initialize DBUS");
+        } else if (dbuscode == DBUS_INIT_ALREADY_RUNNING) {
+                if (radio_url) {
+                        g_debug ("Playing radio URL %s in running instance",
+                                 radio_url);
+                        lastfm_dbus_play_radio_url (radio_url);
+                }
+                return;
+        }
+
+        mainwin = VGL_MAIN_WINDOW (vgl_main_window_new ());
+        g_object_add_weak_pointer (G_OBJECT (mainwin), (gpointer) &mainwin);
+        vgl_main_window_show(mainwin, TRUE);
+        vgl_main_window_set_state(mainwin, VGL_MAIN_WINDOW_STATE_DISCONNECTED,
+                                  NULL);
 
         http_init();
         check_usercfg(FALSE);
         playlist = lastfm_pls_new();
 
-        if (!lastfm_audio_init()) {
+        if (!errmsg && !lastfm_audio_init()) {
                 controller_show_error(_("Error initializing audio system"));
                 return;
         }
-        errmsg = lastfm_dbus_init();
+
+#ifdef MAEMO
+        /* Initialize osso context */
+        if (!errmsg) {
+                osso_context = osso_initialize (APP_NAME_LC,
+                                                APP_VERSION, FALSE, NULL);
+                if (!osso_context) {
+                        errmsg = _("Unable to initialize OSSO context");
+                }
+        }
+#endif
         if (!errmsg) {
                 errmsg = connection_init();
         }
@@ -1501,21 +1682,23 @@ controller_run_app(lastfm_mainwin *win, const char *radio_url)
 
 #ifdef HAVE_TRAY_ICON
         /* Init Freedesktop tray icon */
-        tray_icon = vagalume_tray_icon_create ();
-        vagalume_tray_icon_notify_playback (tray_icon, NULL);
+        tray_icon = vgl_tray_icon_create ();
+        g_object_add_weak_pointer (G_OBJECT (tray_icon), (gpointer) &tray_icon);
+        vgl_tray_icon_notify_playback (tray_icon, NULL);
         if (usercfg != NULL) {
-                vagalume_tray_icon_show_notifications (
+                vgl_tray_icon_show_notifications (
                         tray_icon, usercfg->show_notifications);
         }
 #endif
 
-        mainwin_run_app();
+        vgl_main_window_run_app();
+
+        /* --- From here onwards the app shuts down --- */
 
         lastfm_dbus_notify_closing();
 
 #ifdef HAVE_TRAY_ICON
         g_object_unref(tray_icon);
-        tray_icon = NULL;
 #endif
         lastfm_session_destroy(session);
         session = NULL;
@@ -1523,16 +1706,20 @@ controller_run_app(lastfm_mainwin *win, const char *radio_url)
         rsp_sess = NULL;
         lastfm_pls_destroy(playlist);
         playlist = NULL;
-        lastfm_mainwin_destroy(mainwin);
-        mainwin = NULL;
         if (usercfg != NULL) {
                 set_user_tag_list(usercfg->username, NULL);
                 set_friend_list(usercfg->username, NULL);
-                lastfm_usercfg_destroy(usercfg);
+                vgl_user_cfg_destroy(usercfg);
                 usercfg = NULL;
         }
         lastfm_audio_clear();
         lastfm_dbus_close();
+        vgl_bookmark_mgr_save_to_disk (vgl_bookmark_mgr_get_instance (), FALSE);
+
+#ifdef MAEMO
+        /* Cleanup OSSO context */
+        osso_deinitialize(osso_context);
+#endif
 }
 
 /**
