@@ -15,10 +15,10 @@
 gboolean lastfm_get_friends(const char *username, GList **friendlist);
 gboolean lastfm_get_user_tags(const char *username, GList **taglist);
 gboolean lastfm_get_user_track_tags(const char *username,
-                                    const lastfm_track *track,
+                                    const LastfmTrack *track,
                                     request_type req, GList **taglist);
-gboolean lastfm_get_track_tags(const lastfm_track *track,
+gboolean lastfm_get_track_tags(const LastfmTrack *track,
                                request_type req, GList **taglist);
-void lastfm_get_track_cover_image(lastfm_track *track);
+void lastfm_get_track_cover_image(LastfmTrack *track);
 
 #endif
