@@ -22,6 +22,10 @@ static const char *custom_pls_path =
        "/1.0/webclient/getresourceplaylist.php";
 static const xmlChar *free_track_rel = (xmlChar *)
        "http://www.last.fm/freeTrackURL";
+static const xmlChar *album_page_rel = (xmlChar *)
+       "http://www.last.fm/albumpage";
+static const char *lastfm_music_prefix =
+       "http://www.last.fm/music/";
 
 /**
  * Get the 2-letter language code from the currently active language
@@ -80,29 +84,29 @@ lastfm_parse_handshake(const char *buffer)
 }
 
 /**
- * Destroy a lastfm_session object, freeing all of its memory
+ * Destroy a LastfmSession object, freeing all of its memory
  * @param session The session
  */
 void
-lastfm_session_destroy(lastfm_session *session)
+lastfm_session_destroy(LastfmSession *session)
 {
         if (session == NULL) return;
         g_free(session->id);
         g_free(session->base_url);
         g_free(session->base_path);
-        g_slice_free(lastfm_session, session);
+        g_slice_free(LastfmSession, session);
 }
 
 /**
- * Copy a lastfm_session object
+ * Copy a LastfmSession object
  * @param session The original object
  * @return A new object
  */
-lastfm_session *
-lastfm_session_copy(const lastfm_session *session)
+LastfmSession *
+lastfm_session_copy(const LastfmSession *session)
 {
         if (session == NULL) return NULL;
-        lastfm_session *s = g_slice_new0(lastfm_session);
+        LastfmSession *s = g_slice_new0(LastfmSession);
         *s = *session;
         s->id = g_strdup(session->id);
         s->base_url = g_strdup(session->base_url);
@@ -115,11 +119,11 @@ lastfm_session_copy(const lastfm_session *session)
  * @param username User's ID
  * @param password User's passwords
  * @param err If non-NULL, an error code will be written here
- * @return A new lastfm_session or NULL if it couldn't be created
+ * @return A new LastfmSession or NULL if it couldn't be created
  */
-lastfm_session *
+LastfmSession *
 lastfm_session_new(const char *username, const char *password,
-                   lastfm_err *err)
+                   LastfmErr *err)
 {
         g_return_val_if_fail(username != NULL && password != NULL, NULL);
         char *buffer = NULL;
@@ -140,7 +144,7 @@ lastfm_session_new(const char *username, const char *password,
 
         g_free(buffer);
 
-        lastfm_session *s = g_slice_new0(lastfm_session);
+        LastfmSession *s = g_slice_new0(LastfmSession);
         s->id = g_strdup(g_hash_table_lookup(response, "session"));
         s->base_url = g_strdup(g_hash_table_lookup(response, "base_url"));
         s->base_path = g_strdup(g_hash_table_lookup(response, "base_path"));
@@ -169,7 +173,7 @@ lastfm_session_new(const char *username, const char *password,
  * @return Whether a track has been found and added to the playlist
  */
 static gboolean
-lastfm_parse_track(xmlDoc *doc, xmlNode *node, lastfm_pls *pls,
+lastfm_parse_track(xmlDoc *doc, xmlNode *node, LastfmPls *pls,
                    const char *pls_title, gboolean custom_pls)
 {
         g_return_val_if_fail(doc!=NULL && node!=NULL && pls!=NULL, FALSE);
@@ -177,7 +181,7 @@ lastfm_parse_track(xmlDoc *doc, xmlNode *node, lastfm_pls *pls,
         gboolean retval = FALSE;
         const xmlChar *name;
         char *val;
-        lastfm_track *track = lastfm_track_new();
+        LastfmTrack *track = lastfm_track_new();
         track->pls_title = g_strdup(pls_title);
         track->custom_pls = custom_pls;
 
@@ -194,6 +198,8 @@ lastfm_parse_track(xmlDoc *doc, xmlNode *node, lastfm_pls *pls,
                         track->title = g_strstrip(g_strdup(val));
                 } else if (!xmlStrcmp(name, (xmlChar *) "id")) {
                         track->id = strtol(val, NULL, 10);
+                } else if (!xmlStrcmp(name, (xmlChar *) "artistId")) {
+                        track->artistid = strtol(val, NULL, 10);
                 } else if (!xmlStrcmp(name, (xmlChar *) "creator")) {
                         track->artist = g_strstrip(g_strdup(val));
                 } else if (!xmlStrcmp(name, (xmlChar *) "album")) {
@@ -210,6 +216,14 @@ lastfm_parse_track(xmlDoc *doc, xmlNode *node, lastfm_pls *pls,
                                 if (!xmlStrcmp(rel, free_track_rel)) {
                                         track->free_track_url =
                                                 g_strstrip(g_strdup(val));
+                                } else if (!xmlStrcmp(rel, album_page_rel) &&
+                                           g_str_has_prefix (
+                                                   val, lastfm_music_prefix)) {
+                                        char *artist, **parts;
+                                        parts = g_strsplit (val, "/", 6);
+                                        artist = lastfm_url_decode (parts[4]);
+                                        g_strfreev (parts);
+                                        track->album_artist = artist;
                                 }
                                 xmlFree(rel);
                         }
@@ -224,8 +238,20 @@ lastfm_parse_track(xmlDoc *doc, xmlNode *node, lastfm_pls *pls,
         } else if (track->artist == NULL || track->artist[0] == '\0') {
                 g_debug("Found track with no artist, discarding it");
         } else {
-                if (track->album == NULL) track->album = g_strdup("");
+                if (track->album == NULL) {
+                        /* album_artist != NULL unlikely if album == NULL,
+                           but we free it just in case */
+                        g_free ((gpointer) track->album_artist);
+                        track->album = g_strdup("");
+                        track->album_artist = g_strdup("");
+                }
                 if (track->trackauth == NULL) track->trackauth = g_strdup("");
+                if (track->album_artist == NULL ||
+                    !strcmp (track->artist, track->album_artist)) {
+                        /* Don't waste memory */
+                        g_free ((gpointer) track->album_artist);
+                        track->album_artist = track->artist;
+                }
                 lastfm_pls_add_track(pls, track);
                 retval = TRUE;
         }
@@ -239,13 +265,13 @@ lastfm_parse_track(xmlDoc *doc, xmlNode *node, lastfm_pls *pls,
  * @param bufsize Size of the buffer
  * @return A new playlist, or NULL if none was found
  */
-static lastfm_pls *
+static LastfmPls *
 lastfm_parse_playlist(const char *buffer, size_t bufsize)
 {
         xmlDoc *doc = NULL;
         xmlNode *node = NULL;
         xmlNode *tracklist = NULL;
-        lastfm_pls *pls = NULL;
+        LastfmPls *pls = NULL;
         const xmlChar *name;
         char *pls_title = NULL;
         gboolean custom_pls = FALSE;
@@ -269,17 +295,8 @@ lastfm_parse_playlist(const char *buffer, size_t bufsize)
                         char *title = (char *) xmlNodeListGetString(doc,
                                                node->xmlChildrenNode, 1);
                         if (title != NULL && pls_title == NULL) {
-                                int i;
-                                /* The playlist contains ampersands
-                                 * encoded as '%26amp%3B', so we must
-                                 * decode them in two steps */
-                                char *tmp = escape_url(title, FALSE);
-                                pls_title = string_replace(tmp, "&amp;", "&");
-                                g_free(tmp);
-                                for (i = 0; pls_title[i] != 0; i++) {
-                                        if (pls_title[i] == '+')
-                                                pls_title[i] = ' ';
-                                }
+                                /* The playlist comes encoded */
+                                pls_title = lastfm_url_decode (title);
                         }
                         xmlFree((xmlChar *)title);
                 } else if (!xmlStrcmp(name, (const xmlChar *) "trackList")) {
@@ -318,15 +335,15 @@ lastfm_parse_playlist(const char *buffer, size_t bufsize)
  * @param discovery Whether to use discovery mode or not
  * @return A new playlist or NULL if none has been obtained
  */
-lastfm_pls *
-lastfm_request_playlist(lastfm_session *s, gboolean discovery)
+LastfmPls *
+lastfm_request_playlist(LastfmSession *s, gboolean discovery)
 {
         g_return_val_if_fail(s && s->id && s->base_url && s->base_path, NULL);
         const char *disc_mode = discovery ? "1" : "0";
         char *url;
         char *buffer = NULL;
         size_t bufsize = 0;
-        lastfm_pls *pls = NULL;
+        LastfmPls *pls = NULL;
 
         url = g_strconcat("http://", s->base_url, s->base_path,
                           "/xspf.php?sk=", s->id, "&discovery=", disc_mode,
@@ -347,13 +364,13 @@ lastfm_request_playlist(lastfm_session *s, gboolean discovery)
  * @param radio_url URL of the playlist
  * @return A new playlist, or NULL if none has been obtained
  */
-lastfm_pls *
-lastfm_request_custom_playlist(lastfm_session *s, const char *radio_url)
+LastfmPls *
+lastfm_request_custom_playlist(LastfmSession *s, const char *radio_url)
 {
         g_return_val_if_fail(s != NULL && radio_url != NULL, NULL);
         char *buffer = NULL;
         size_t bufsize = 0;
-        lastfm_pls *pls = NULL;
+        LastfmPls *pls = NULL;
         char *url = NULL;
         char *radio_url_escaped = escape_url(radio_url, TRUE);
         url = g_strconcat("http://", s->base_url, custom_pls_path,
@@ -377,7 +394,7 @@ lastfm_request_custom_playlist(lastfm_session *s, const char *radio_url)
  * @return Whether the radio has been set correctly
  */
 gboolean
-lastfm_set_radio(lastfm_session *s, const char *radio_url)
+lastfm_set_radio(LastfmSession *s, const char *radio_url)
 {
         g_return_val_if_fail(s != NULL && s->id != NULL &&
                              s->base_url != NULL &&
