@@ -14,7 +14,7 @@ typedef enum {
         LASTFM_GROUP_RADIO,
         LASTFM_LOVEDTRACKS_RADIO,
         LASTFM_NEIGHBOURS_RADIO,
-        LASTFM_PERSONAL_RADIO,
+        LASTFM_LIBRARY_RADIO,
         LASTFM_SIMILAR_ARTIST_RADIO,
         LASTFM_RECOMMENDED_RADIO,
         LASTFM_USERPLAYLIST_RADIO,
