@@ -16,10 +16,10 @@
 
 GtkWidget *compat_gtk_button_new(void);
 gboolean tagwin_run(GtkWindow *parent, const char *user, char **newtags,
-                    const GList *usertags, lastfm_track *track,
+                    const GList *usertags, LastfmTrack *track,
                     request_type *type);
 gboolean recommwin_run(GtkWindow *parent, char **user, char **message,
-                       const GList *friends, const lastfm_track *track,
+                       const GList *friends, const LastfmTrack *track,
                        request_type *type);
 void ui_info_banner(GtkWindow *parent, const char *text);
 void ui_info_dialog(GtkWindow *parent, const char *text);
@@ -30,7 +30,9 @@ char *ui_input_dialog(GtkWindow *parent, const char *title,
 char *ui_input_dialog_with_list(GtkWindow *parent, const char *title,
                                 const char *text, const GList *elems,
                                 const char *value);
-gboolean ui_usercfg_window(GtkWindow *parent, lastfm_usercfg **cfg);
+gboolean ui_edit_bookmark_dialog(GtkWindow *parent, char **name,
+                                 char **url, gboolean add);
+gboolean ui_usercfg_window(GtkWindow *parent, VglUserCfg **cfg);
 gboolean ui_confirm_dialog(GtkWindow *parent, const char *text);
 void flush_ui_events(void);
 
